"""Unit testing for the galmatch module."""
import numpy as np
from ..galmatch import galsample
from ..galmatch import calculate_indx_correspondence


def test_source_galaxy_selection_indices():
    """Set up a simple example of source and target halo catalogs with a few tightly
    localized collections of properties. Verify that the bookkeeping of the
    compute_source_galaxy_selection_indices function correctly handles the
    correspondence between halos with similar properties.
    """
    n_groupings = 3
    n_source_halos_per_group, n_target_halos_per_group = 50, 600
    n_source_halos = n_groupings * n_source_halos_per_group
    n_target_halos = n_groupings * n_target_halos_per_group

    source_halo_ids = np.arange(n_source_halos).astype(int)
    target_halo_ids = np.arange(n_target_halos).astype(int)

    source_halos_x = np.zeros(n_source_halos)
    target_halos_x = np.zeros(n_target_halos)

    source_halo_richness = np.zeros(n_source_halos).astype(int)
    richness1, richness2, richness3 = 2, 4, 5
    x1, x2, x3 = 1, 10, -10

    i, j = 0, n_source_halos_per_group
    source_halos_x[i:j] = np.random.normal(
        loc=x1, scale=0.01, size=n_source_halos_per_group
    )
    source_halo_richness[i:j] = richness1
    i, j = n_source_halos_per_group, 2 * n_source_halos_per_group
    source_halos_x[i:j] = np.random.normal(
        loc=x2, scale=0.01, size=n_source_halos_per_group
    )
    source_halo_richness[i:j] = richness2
    i, j = 2 * n_source_halos_per_group, 3 * n_source_halos_per_group
    source_halos_x[i:j] = np.random.normal(
        loc=x3, scale=0.01, size=n_source_halos_per_group
    )
    source_halo_richness[i:j] = richness3

    i, j = 0, n_target_halos_per_group
    target_halos_x[i:j] = np.random.normal(
        loc=x1, scale=0.01, size=n_target_halos_per_group
    )
    i, j = n_target_halos_per_group, 2 * n_target_halos_per_group
    target_halos_x[i:j] = np.random.normal(
        loc=x2, scale=0.01, size=n_target_halos_per_group
    )
    i, j = 2 * n_target_halos_per_group, 3 * n_target_halos_per_group
    target_halos_x[i:j] = np.random.normal(
        loc=x3, scale=0.01, size=n_target_halos_per_group
    )

    source_halo_props = (source_halos_x,)
    target_halo_props = (target_halos_x,)

    source_galaxies_host_halo_id = np.repeat(source_halo_ids, source_halo_richness)
    res = galsample(
        source_galaxies_host_halo_id,
        source_halo_ids,
        target_halo_ids,
        source_halo_props,
        target_halo_props,
    )
    selection_indx = res.target_gals_selection_indx
    target_galaxy_target_halo_ids = res.target_gals_target_halo_ids
    n_mock_correct = n_target_halos_per_group * (richness1 + richness2 + richness3)
    assert selection_indx.size == n_mock_correct

    source_gals_host_x = np.repeat(source_halos_x, source_halo_richness)
    target_gals_source_host_x = source_gals_host_x[selection_indx]

    # Check that the mock has inherited from the right halos
    msk1 = target_galaxy_target_halo_ids < n_target_halos_per_group
    dx1 = np.abs(target_gals_source_host_x[msk1] - x1)
    assert np.all(dx1 < 1)

    assert ~np.all(np.abs(target_gals_source_host_x - x1) < 1)

    msk2 = ~msk1 & (target_galaxy_target_halo_ids < 2 * n_target_halos_per_group)
    dx2 = np.abs(target_gals_source_host_x[msk2] - x2)
    assert np.all(dx2 < 1)


def test_calculate_indx_correspondence():
    n_source, n_target_per_source = 10, 5
    n_target = n_source * n_target_per_source
    x_source = np.arange(n_source)

    delta = 0.01
    u_target = np.random.uniform(-delta, delta, n_target)
    x_target = np.repeat(x_source, n_target_per_source) + u_target
    source_props = (x_source,)
    target_props = (x_target,)
    dd_match, indx_match = calculate_indx_correspondence(source_props, target_props)
    assert dd_match.shape == x_target.shape
    assert dd_match.max() <= delta

    x_match = x_source[indx_match]
    assert np.allclose(x_target, x_match, atol=delta)



"""
"""
import numpy as np


__all__ = ("crossmatch", "compute_richness")


def crossmatch(x, y, skip_bounds_checking=False):
    """
    Finds where the elements of ``x`` appear in the array ``y``, including repeats.

    The elements in x may be repeated, but the elements in y must be unique.
    The arrays x and y may be only partially overlapping.

    Applications of this function involve cross-matching two catalogs which share an ID.

    Parameters
    ----------
    x : integer array
        Array of integers with possibly repeated entries.

    y : integer array
        Array of unique integers.

    skip_bounds_checking : bool, optional
        The first step in the `crossmatch` function is to test that the input
        arrays satisfy the assumptions of the algorithm
        (namely that ``x`` and ``y`` store integers,
        and that all values in ``y`` are unique).
        If ``skip_bounds_checking`` is set to True,
        this testing is bypassed and the function evaluates faster.
        Default is False.

    Returns
    -------
    idx_x : integer array
        Integer array used to apply a mask to x
        such that x[idx_x] = y[idx_y]

    y_idx : integer array
        Integer array used to apply a mask to y
        such that x[idx_x] = y[idx_y]

    """
    # Ensure inputs are Numpy arrays
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)

    # Require that the inputs meet the assumptions of the algorithm
    if skip_bounds_checking is True:
        pass
    else:
        try:
            assert len(set(y)) == len(y)
            assert np.all(np.array(y, dtype=np.int64) == y)
            assert np.shape(y) == (len(y),)
        except (AssertionError, ValueError, TypeError):
            msg = "Input array y must be a 1d sequence of unique integers"
            raise ValueError(msg)
        try:
            assert np.all(np.array(x, dtype=np.int64) == x)
            assert np.shape(x) == (len(x),)
        except (AssertionError, ValueError, TypeError):
            msg = "Input array x must be a 1d sequence of integers"
            raise ValueError(msg)

    # Internally, we will work with sorted arrays, and then undo the sorting at the end
    idx_x_sorted = np.argsort(x)
    idx_y_sorted = np.argsort(y)
    x_sorted = np.copy(x[idx_x_sorted])
    y_sorted = np.copy(y[idx_y_sorted])

    # x may have repeated entries
    # Address by finding the unique values as well as their multiplicity
    unique_xvals, counts = np.unique(x_sorted, return_counts=True)

    # Determine which of the unique x values has a match in y
    unique_xval_has_match = np.in1d(unique_xvals, y_sorted, assume_unique=True)

    # Create a boolean array with True for each value in x with a match, otherwise False
    idx_x = np.repeat(unique_xval_has_match, counts)

    # For each unique value of x with a match in y, identify the index of the match
    matching_indices_in_y = np.searchsorted(
        y_sorted, unique_xvals[unique_xval_has_match]
    )

    # Repeat each matching index according to the multiplicity in x
    idx_y = np.repeat(matching_indices_in_y, counts[unique_xval_has_match])

    # Undo the original sorting and return the result
    return idx_x_sorted[idx_x], idx_y_sorted[idx_y]


def compute_richness(unique_halo_ids, halo_id_of_galaxies):
    r"""For every ID in unique_halo_ids,
    calculate the number of times the ID appears in halo_id_of_galaxies.

    Parameters
    ----------
    unique_halo_ids : ndarray
        Numpy array of shape (num_halos, ) storing unique integers

    halo_id_of_galaxies : ndarray
        Numpy integer array of shape (num_galaxies, ) storing the host ID of each galaxy

    Returns
    -------
    richness : ndarray
        Numpy integer array of shape (num_halos, ) storing richness of each host halo

    Examples
    --------
    >>> num_hosts = 100
    >>> num_sats = int(1e5)
    >>> unique_halo_ids = np.arange(5, num_hosts + 5)
    >>> halo_id_of_galaxies = np.random.randint(0, 5000, num_sats)
    >>> richness = compute_richness(unique_halo_ids, halo_id_of_galaxies)
    """
    unique_halo_ids = np.atleast_1d(unique_halo_ids).astype(int)
    halo_id_of_galaxies = np.atleast_1d(halo_id_of_galaxies).astype(int)
    richness_result = np.zeros_like(unique_halo_ids).astype(int)

    vals, counts = np.unique(halo_id_of_galaxies, return_counts=True)
    idxA, idxB = crossmatch(vals, unique_halo_ids)
    richness_result[idxB] = counts[idxA]
    return richness_result



""" Module providing unit-testing of `~halotools.utils.crossmatch` function.
"""
import numpy as np
import pytest

from ..crossmatch import crossmatch

fixed_seed = 43


def test_crossmatch1():
    """x has unique entries. All y values are in x. All x values are in y."""
    x = np.array([1, 3, 5])
    y = np.array([5, 1])
    x_idx, y_idx = crossmatch(x, y)

    assert np.all(x[x_idx] == y[y_idx])


def test_crossmatch2():
    """x has repeated entries. All y values are in x. All x values are in y."""
    x = np.array([1, 3, 5, 3, 1, 1, 3, 5])
    y = np.array([5, 1])
    x_idx, y_idx = crossmatch(x, y)

    assert np.all(x[x_idx] == y[y_idx])


def test_crossmatch3():
    """x has repeated entries. All y values are in x. Some x values are not in y."""
    x = np.array([0, 1, 3, 5, 3, -1, 1, 3, 5, -1])
    y = np.array([5, 1])
    x_idx, y_idx = crossmatch(x, y)

    assert np.all(x[x_idx] == y[y_idx])


def test_crossmatch4():
    """x has repeated entries. Some y values are not in x. Some x values are not in y"""
    x = np.array([1, 3, 5, 3, 1, -1, 3, 5, -10, -10])
    y = np.array([5, 1, 100, 20])
    x_idx, y_idx = crossmatch(x, y)

    assert np.all(x[x_idx] == y[y_idx])


def test_crossmatch5():
    """x has repeated entries. Some y values are not in x. Some x values are not in y"""
    xmax = 100
    numx = 10000
    rng = np.random.RandomState(fixed_seed)
    x = rng.randint(0, xmax + 1, numx)

    y = np.arange(-xmax, xmax)[::10]
    rng = np.random.RandomState(fixed_seed + 1)
    rng.shuffle(y)

    x_idx, y_idx = crossmatch(x, y)

    assert np.all(x[x_idx] == y[y_idx])


def test_crossmatch6():
    """x and y have zero overlap."""
    x = np.array([-1, -5, -10])
    y = np.array([1, 2, 3, 4])
    x_idx, y_idx = crossmatch(x, y)
    assert len(x_idx) == 0
    assert len(y_idx) == 0
    assert np.all(x[x_idx] == y[y_idx])


def test_error_handling1():
    """Verify that we raise the proper exception when y has repeated entries."""
    x = np.ones(5)
    y = np.ones(5)

    with pytest.raises(ValueError) as err:
        crossmatch(x, y)
    substr = "Input array y must be a 1d sequence of unique integers"
    assert substr in err.value.args[0]


def test_error_handling2():
    """Verify that we raise the proper exception when y has non-integer values."""
    x = np.ones(5)
    y = np.arange(0, 5, 0.5)

    with pytest.raises(ValueError) as err:
        crossmatch(x, y)
    substr = "Input array y must be a 1d sequence of unique integers"
    assert substr in err.value.args[0]


def test_error_handling3():
    """Verify that we raise the proper exception when y is multi-dimensional."""
    x = np.ones(5)
    y = np.arange(0, 6).reshape(2, 3)

    with pytest.raises(ValueError) as err:
        crossmatch(x, y)
    substr = "Input array y must be a 1d sequence of unique integers"
    assert substr in err.value.args[0]


def test_error_handling4():
    """Verify that we raise the proper exception when x has non-integer values."""
    x = np.arange(0, 5, 0.5)
    y = np.arange(0, 6)

    with pytest.raises(ValueError) as err:
        crossmatch(x, y)
    substr = "Input array x must be a 1d sequence of integers"
    assert substr in err.value.args[0]


def test_error_handling5():
    """Verify that we raise the proper exception when x is multi-dimensional."""
    x = np.arange(0, 6).reshape(2, 3)
    y = np.arange(0, 6)

    with pytest.raises(ValueError) as err:
        crossmatch(x, y)
    substr = "Input array x must be a 1d sequence of integers"
    assert substr in err.value.args[0]


def test_crossmatch_skip_bounds_checking():
    """skip_bounds_checking should not change results."""
    x = np.array([1, 3, 5])
    y = np.array([5, 1])
    x_idx, y_idx = crossmatch(x, y, skip_bounds_checking=True)
    x_idx2, y_idx2 = crossmatch(x, y, skip_bounds_checking=False)

    assert np.allclose(x_idx, x_idx2)
    assert np.allclose(y_idx, y_idx2)



# Configuration file for the Sphinx documentation builder.
#
"""
"""
from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("galsampler").version
except DistributionNotFound:
    __version__ = "unknown version"

project = "galsampler"
copyright = "2022, Andrew Hearin"
author = "Andrew Hearin"
version = __version__
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "nbsphinx",
]

templates_path = ["_templates"]
exclude_patterns = [".ipynb_checkpoints/*"]

nbsphinx_prompt_width = "0"  # no prompts in nbsphinx

nbsphinx_kernel_name = "python3"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = []

master_doc = "index"



"""Module implementing the galsample function
used for galsampling
"""
import numpy as np
from scipy.spatial import cKDTree
from numba import njit
from collections import namedtuple
from .crossmatch import crossmatch, compute_richness

GalsamplerCorrespondence = namedtuple(
    "GalsamplerCorrespondence",
    [
        "target_gals_selection_indx",
        "target_gals_target_halo_ids",
        "target_gals_source_halo_ids",
    ],
)

__all__ = ("galsample", "calculate_halo_correspondence")


@njit
def galaxy_selection_kernel(first_source_gal_indices, richness, n_target_halo, result):
    """Numba kernel filling in array of galaxy selection indices

    Parameters
    ----------
    first_source_gal_indices : ndarray of shape (n_target_halo, )
        Stores the index of the first galaxy in the galaxy catalog
        assigned to each target halo

    richness : ndarray of shape (n_target_halo, )
        Stores the number of galaxies that will be mapped to each target halo

    n_target_halo : int

    result : ndarray of shape richness.sum()

    """
    cur = 0
    for i in range(n_target_halo):
        ifirst = first_source_gal_indices[i]
        n = richness[i]
        if n > 0:
            ilast = ifirst + richness[i]
            for j in range(ifirst, ilast):
                result[cur] = j
                cur += 1


def _get_data_block(*halo_properties):
    return np.vstack(halo_properties).T


def calculate_halo_correspondence(source_halo_props, target_halo_props, n_threads=-1):
    """Calculating indexing array defined by a statistical correspondence between
    source and target halos.

    Parameters
    ----------
    source_halo_props : sequence of n_props ndarrays
        Each ndarray should have shape (n_source_halos, )

    target_halo_props : sequence of n_props ndarrays
        Each ndarray should have shape (n_target_halos, )

    Returns
    -------
    dd_match : ndarray of shape (n_target_halos, )
        Euclidean distance to the source halo matched to each target halo

    indx_match : ndarray of shape (n_target_halos, )
        Index of the source halo matched to each target halo

    """
    assert len(source_halo_props) == len(target_halo_props)
    X_source = _get_data_block(*source_halo_props)
    X_target = _get_data_block(*target_halo_props)
    source_tree = cKDTree(X_source)
    dd_match, indx_match = source_tree.query(X_target, workers=n_threads)
    return dd_match, indx_match


def galsample(
    source_galaxies_host_halo_id,
    source_halo_ids,
    target_halo_ids,
    source_halo_props,
    target_halo_props,
):
    """Calculate the indexing array that transfers source galaxies to target halos

    Parameters
    ----------
    source_galaxies_host_halo_id : ndarray of shape (n_source_gals, )
        Integer array storing the value of the halo ID
        of the source halo of each source galaxy
        Each entry in source_galaxies_host_halo_id must appear in source_halo_ids
        This quantity can be computed via the following function in halotools:
        halotools.utils.value_added_halo_table_functions.compute_uber_hostid

    source_halo_ids : ndarray of shape (n_source_halos, )

    target_halo_ids : ndarray of shape (n_target_halos, )

    source_halo_props : sequence of n_props ndarrays
        Sequence of n_props of ndarrays, each with shape (n_source_halos, )
        These properties determine the correspondence between source and target halos

    target_halo_props : sequence of n_props ndarrays
        Sequence of n_props of ndarrays, each with shape (n_target_halos, )
        These properties determine the correspondence between source and target halos

    Returns
    -------
    target_gals_selection_indx : ndarray of shape (n_target_gals, )
        Integer array storing values in the range [0, n_source_gals-1]

    target_galaxy_target_halo_ids : ndarray of shape (n_target_gals, )
        Integer array storing values appearing in target_halo_ids

    target_galaxy_source_halo_ids : ndarray of shape (n_target_gals, )
        Integer array storing values appearing in source_halo_ids

    """
    #  Sort the source galaxies so that members of a common halo are grouped together
    idx_sorted_source_galaxies = np.argsort(source_galaxies_host_halo_id)
    sorted_source_galaxies_host_halo_id = source_galaxies_host_halo_id[
        idx_sorted_source_galaxies
    ]

    #  Calculate the index correspondence array that will undo the sorting at the end
    num_source_gals = len(source_galaxies_host_halo_id)
    idx_unsorted_galaxy_indices = np.arange(num_source_gals).astype("i8")[
        idx_sorted_source_galaxies
    ]

    #  For each source halo, calculate the number of resident galaxies
    source_halos_richness = compute_richness(
        source_halo_ids, sorted_source_galaxies_host_halo_id
    )

    #  For each source halo, calculate the index of its first resident galaxy
    source_halo_sorted_source_galaxies_indices = _galaxy_table_indices(
        source_halo_ids, sorted_source_galaxies_host_halo_id
    )

    #  For each target halo, calculate the index of the associated source halo
    __, source_halo_selection_indices = calculate_halo_correspondence(
        source_halo_props, target_halo_props
    )

    #  For each target halo, calculate the number of galaxies
    target_halo_richness = source_halos_richness[source_halo_selection_indices]
    num_target_gals = np.sum(target_halo_richness)

    #  For each target halo, calculate the halo ID of the associated source halo
    target_halo_source_halo_ids = source_halo_ids[source_halo_selection_indices]

    #  For each target halo, calculate the index of its first resident galaxy
    target_halo_first_sorted_source_gal_indices = (
        source_halo_sorted_source_galaxies_indices[source_halo_selection_indices]
    )

    #  For every target halo, we know the index of the first and last galaxy to select
    #  Calculate an array of shape (num_target_gals, )
    # with the index of each selected galaxy
    n_target_halos = target_halo_ids.size
    sorted_source_galaxy_selection_indices = np.zeros(num_target_gals).astype(int)
    galaxy_selection_kernel(
        target_halo_first_sorted_source_gal_indices.astype("i8"),
        target_halo_richness.astype("i4"),
        n_target_halos,
        sorted_source_galaxy_selection_indices,
    )

    #  For each target galaxy, calculate the halo ID of its source and target halo
    target_gals_target_halo_ids = np.repeat(target_halo_ids, target_halo_richness)
    target_gals_source_halo_ids = np.repeat(
        target_halo_source_halo_ids, target_halo_richness
    )

    #  For each index in the sorted galaxy catalog,
    #  calculate the index of the catalog in its original order
    target_gals_selection_indx = idx_unsorted_galaxy_indices[
        sorted_source_galaxy_selection_indices
    ]

    return GalsamplerCorrespondence(
        target_gals_selection_indx,
        target_gals_target_halo_ids,
        target_gals_source_halo_ids,
    )


def _galaxy_table_indices(source_halo_id, galaxy_host_halo_id):
    """For every halo in the source halo catalog, calculate the index
    in the source galaxy catalog of the first appearance of a galaxy that
    occupies the halo, reserving -1 for source halos with no resident galaxies.

    Parameters
    ----------
    source_halo_id : ndarray
        Numpy integer array of shape (num_halos, )

    galaxy_host_halo_id : ndarray
        Numpy integer array of shape (num_gals, )

    Returns
    -------
    indices : ndarray
        Numpy integer array of shape (num_halos, ).
        All values will be in the interval [-1, num_gals)
    """
    uval_gals, indx_uval_gals = np.unique(galaxy_host_halo_id, return_index=True)
    idxA, idxB = crossmatch(source_halo_id, uval_gals)
    num_source_halos = len(source_halo_id)
    indices = np.zeros(num_source_halos) - 1
    indices[idxA] = indx_uval_gals[idxB]
    return indices.astype(int)


def calculate_indx_correspondence(source_props, target_props, n_threads=-1):
    """For each target data object, find a closely matching source data object

    Parameters
    ----------
    source_props : list of n_props ndarrays
        Each ndarray should have shape (n_source, )

    target_props : list of n_props ndarrays
        Each ndarray should have shape (n_target, )

    Returns
    -------
    dd_match : ndarray of shape (n_target, )
        Euclidean distance between each target and its matching source object

    indx_match : ndarray of shape (n_target, )
        Index into the source object that is matched to each target

    Notes
    -----
    This function essentially provides wrapper behavior around scipy.spatial.cKDTree

    """
    assert len(source_props) == len(target_props)
    X_source = _get_data_block(*source_props)
    X_target = _get_data_block(*target_props)
    source_tree = cKDTree(X_source)
    dd_match, indx_match = source_tree.query(X_target, workers=n_threads)
    return dd_match, indx_match
